
#ifndef NTL_ZZXY__H
#define NTL_ZZXY__H

#include <NTL/vec_ZZX.h>
#include <NTL/lzz_pX.h>
#include <NTL/ZZ_pX.h>

NTL_OPEN_NNS


class ZZXY {

public:

vec_ZZX rep;


/***************************************************************

          Constructors, Destructors, and Assignment

****************************************************************/


ZZXY()
//  initial value 0

   { }


ZZXY(INIT_SIZE_TYPE, long n)
// initial value 0, but space is pre-allocated for n coefficients

   { rep.SetMaxLength(n); }

ZZXY(const ZZXY& a) : rep(a.rep) { }
// initial value is a


ZZXY& operator=(const ZZXY& a)
   { rep = a.rep; return *this; }

~ZZXY() { }

void normalize();
// strip leading zeros

void SetMaxLength(long n)
// pre-allocate space for n coefficients.
// Value is unchanged

   { rep.SetMaxLength(n); }


void kill()
// free space held by this polynomial.  Value becomes 0.

   { rep.kill(); }

static const ZZXY& zero();

inline ZZXY(long i, const ZZX& c);
inline ZZXY(long i, long c);


inline ZZXY& operator=(long a);
inline ZZXY& operator=(const ZZX& a);


ZZXY(ZZXY& x, INIT_TRANS_TYPE) : rep(x.rep, INIT_TRANS) { }

};




/********************************************************************

                           input and output

I/O format:

   [a_0 a_1 ... a_n],

represents the polynomial a_0 + a_1*X + ... + a_n*X^n.

On output, all coefficients will be integers between 0 and p-1,
amd a_n not zero (the zero polynomial is [ ]).
Leading zeroes are stripped.

*********************************************************************/


NTL_SNS istream& operator>>(NTL_SNS istream& s, ZZXY& x);
NTL_SNS ostream& operator<<(NTL_SNS ostream& s, const ZZXY& a);




/**********************************************************

                   Some utility routines

***********************************************************/


inline long deg(const ZZXY& a) { return a.rep.length() - 1; }
// degree of a polynomial.
// note that the zero polynomial has degree -1.

const ZZX& coeff(const ZZXY& a, long i);
// zero if i not in range

void GetCoeff(ZZX& x, const ZZXY& a, long i);
// x = a[i], or zero if i not in range

const ZZX& LeadCoeff(const ZZXY& a);
// zero if a == 0

const ZZX& ConstTerm(const ZZXY& a);
// zero if a == 0

void SetCoeff(ZZXY& x, long i, const ZZX& a);
// x[i] = a, error is raised if i < 0

//void SetCoeff(ZZXY& x, long i, long a);

inline ZZXY::ZZXY(long i, const ZZX& a)
   { SetCoeff(*this, i, a); }

//inline ZZX::ZZX(long i, long a)
  // { SetCoeff(*this, i, a); }

void SetCoeff(ZZXY& x, long i);
// x[i] = 1, error is raised if i < 0

void SetX(ZZXY& x);
// x is set to the monomial X

long IsY(const ZZXY& a);
// test if x = X

inline void clear(ZZXY& x)
// x = 0

   { x.rep.SetLength(0); }

inline void set(ZZXY& x)
// x = 1

   { x.rep.SetLength(1); set(x.rep[0]); }

inline void swap(ZZXY& x, ZZXY& y)
// swap x & y (only pointers are swapped)

   { swap(x.rep, y.rep); }

void trunc(ZZXY& x, const ZZXY& a, long m);
// x = a % X^m

inline ZZXY trunc(const ZZXY& a, long m)
   { ZZXY x; trunc(x, a, m); NTL_OPT_RETURN(ZZXY, x); }

void RightShift(ZZXY& x, const ZZXY& a, long n);
// x = a/X^n

inline ZZXY RightShift(const ZZXY& a, long n)
   { ZZXY x; RightShift(x, a, n); NTL_OPT_RETURN(ZZXY, x); }

void LeftShift(ZZXY& x, const ZZXY& a, long n);
// x = a*X^n

inline ZZXY LeftShift(const ZZXY& a, long n)
   { ZZXY x; LeftShift(x, a, n); NTL_OPT_RETURN(ZZXY, x); }


#ifndef NTL_TRANSITION

inline ZZXY operator>>(const ZZXY& a, long n)
   { ZZXY x; RightShift(x, a, n); NTL_OPT_RETURN(ZZXY, x); }

inline ZZXY operator<<(const ZZXY& a, long n)
   { ZZXY x; LeftShift(x, a, n); NTL_OPT_RETURN(ZZXY, x); }

inline ZZXY& operator<<=(ZZXY& x, long n)
   { LeftShift(x, x, n); return x; }

inline ZZXY& operator>>=(ZZXY& x, long n)
   { RightShift(x, x, n); return x; }

#endif



void reverse(ZZXY& c, const ZZXY& a, long hi);

inline ZZXY reverse(const ZZXY& a, long hi)
   { ZZXY x; reverse(x, a, hi); NTL_OPT_RETURN(ZZXY, x); }

inline void reverse(ZZXY& c, const ZZXY& a)
{  reverse(c, a, deg(a)); }

inline ZZXY reverse(const ZZXY& a)
   { ZZXY x; reverse(x, a); NTL_OPT_RETURN(ZZXY, x); }

//doplneni nebo useknuti na danou delku
inline void VectorCopy(vec_ZZX& x, const ZZXY& a, long n)
   { VectorCopy(x, a.rep, n); }

inline vec_ZZX VectorCopy(const ZZXY& a, long n)
   { return VectorCopy(a.rep, n); }







/*******************************************************************

                        conversion routines

********************************************************************/


void conv(ZZXY& x, long a);
inline ZZXY to_ZZXY(long a)
   { ZZXY x; conv(x, a); NTL_OPT_RETURN(ZZXY, x); }

inline ZZXY& ZZXY::operator=(long a)
   { conv(*this, a); return *this; }

void conv(ZZXY& x, const ZZX& a);
inline ZZXY to_ZZXY(const ZZX& a)
   { ZZXY x; conv(x, a); NTL_OPT_RETURN(ZZXY, x); }

inline ZZXY& ZZXY::operator=(const ZZX& a)
   { conv(*this, a); return *this; }

void conv(ZZXY& x, const vec_ZZX& a);
inline ZZXY to_ZZXY(const vec_ZZX& a)
   { ZZXY x; conv(x, a); NTL_OPT_RETURN(ZZXY, x); }


/*************************************************************

                        Comparison

**************************************************************/

long IsZero(const ZZXY& a);

long IsOne(const ZZXY& a);

long operator==(const ZZXY& a, const ZZXY& b);

inline long operator!=(const ZZXY& a, const ZZXY& b) { return !(a == b); }

long operator==(const ZZXY& a, const ZZX& b);
long operator==(const ZZXY& a, long b);

inline long operator==(const ZZX& a, const ZZXY& b) { return b == a; }
inline long operator==(long a, const ZZXY& b) { return b == a; }

inline long operator!=(const ZZXY& a, const ZZX& b) { return !(a == b); }
inline long operator!=(const ZZXY& a, long b) { return !(a == b); }
inline long operator!=(const ZZX& a, const ZZXY& b) { return !(a == b); }
inline long operator!=(long a, const ZZXY& b) { return !(a == b); }


/***************************************************************

                         Addition

****************************************************************/

void add(ZZXY& x, const ZZXY& a, const ZZXY& b);
// x = a + b

void sub(ZZXY& x, const ZZXY& a, const ZZXY& b);
// x = a - b

void negate(ZZXY& x, const ZZXY& a);
// x = -a

// scalar versions


void add(ZZXY & x, const ZZXY& a, const ZZX& b); // x = a + b
void add(ZZXY& x, const ZZXY& a, long b);

inline void add(ZZXY& x, const ZZX& a, const ZZXY& b) { add(x, b, a); }
inline void add(ZZXY& x, long a, const ZZXY& b) { add(x, b, a); }


void sub(ZZXY & x, const ZZXY& a, const ZZX& b); // x = a - b
void sub(ZZXY& x, const ZZXY& a, long b);

void sub(ZZXY& x, const ZZX& a, const ZZXY& b);
void sub(ZZXY& x, long a, const ZZXY& b);


inline ZZXY operator+(const ZZXY& a, const ZZXY& b)
   { ZZXY x; add(x, a, b); NTL_OPT_RETURN(ZZXY, x); }

inline ZZXY operator+(const ZZXY& a, const ZZX& b)
   { ZZXY x; add(x, a, b); NTL_OPT_RETURN(ZZXY, x); }

inline ZZXY operator+(const ZZXY& a, long b)
   { ZZXY x; add(x, a, b); NTL_OPT_RETURN(ZZXY, x); }

inline ZZXY operator+(const ZZX& a, const ZZXY& b)
   { ZZXY x; add(x, a, b); NTL_OPT_RETURN(ZZXY, x); }

inline ZZXY operator+(long a, const ZZXY& b)
   { ZZXY x; add(x, a, b); NTL_OPT_RETURN(ZZXY, x); }


inline ZZXY operator-(const ZZXY& a, const ZZXY& b)
   { ZZXY x; sub(x, a, b); NTL_OPT_RETURN(ZZXY, x); }

inline ZZXY operator-(const ZZXY& a, const ZZX& b)
   { ZZXY x; sub(x, a, b); NTL_OPT_RETURN(ZZXY, x); }

inline ZZXY operator-(const ZZXY& a, long b)
   { ZZXY x; sub(x, a, b); NTL_OPT_RETURN(ZZXY, x); }

inline ZZXY operator-(const ZZX& a, const ZZXY& b)
   { ZZXY x; sub(x, a, b); NTL_OPT_RETURN(ZZXY, x); }

inline ZZXY operator-(long a, const ZZXY& b)
   { ZZXY x; sub(x, a, b); NTL_OPT_RETURN(ZZXY, x); }


inline ZZXY& operator+=(ZZXY& x, const ZZXY& b)
   { add(x, x, b); return x; }

inline ZZXY& operator+=(ZZXY& x, const ZZX& b)
   { add(x, x, b); return x; }

inline ZZXY& operator+=(ZZXY& x, long b)
   { add(x, x, b); return x; }

inline ZZXY& operator-=(ZZXY& x, const ZZXY& b)
   { sub(x, x, b); return x; }

inline ZZXY& operator-=(ZZXY& x, const ZZX& b)
   { sub(x, x, b); return x; }

inline ZZXY& operator-=(ZZXY& x, long b)
   { sub(x, x, b); return x; }


inline ZZXY operator-(const ZZXY& a)
   { ZZXY x; negate(x, a); NTL_OPT_RETURN(ZZXY, x); }

inline ZZXY& operator++(ZZXY& x) { add(x, x, 1); return x; }
inline void operator++(ZZXY& x, int) { add(x, x, 1); }
inline ZZXY& operator--(ZZXY& x) { sub(x, x, 1); return x; }
inline void operator--(ZZXY& x, int) { sub(x, x, 1); }


/*****************************************************************

                        Multiplication

******************************************************************/


void mul(ZZXY& x, const ZZXY& a, const ZZXY& b);
// x = a * b

void mul(ZZXY& x, const ZZXY& a, const ZZX& b);
void mul(ZZXY& x, const ZZXY& a, const ZZ& b);
void mul(ZZXY& x, const ZZXY& a, long b);

inline void mul(ZZXY& x, const ZZX& a, const ZZXY& b) { mul(x, b, a); }
inline void mul(ZZXY& x, const ZZ& a, const ZZXY& b) { mul(x, b, a);}
inline void mul(ZZXY& x, long a, const ZZXY& b) { mul(x, b, a); }


inline ZZXY operator*(const ZZXY& a, const ZZXY& b)
   { ZZXY x; mul(x, a, b); NTL_OPT_RETURN(ZZXY, x); }

inline ZZXY operator*(const ZZXY& a, const ZZX& b)
   { ZZXY x; mul(x, a, b); NTL_OPT_RETURN(ZZXY, x); }

inline ZZXY operator*(const ZZXY& a, long b)
   { ZZXY x; mul(x, a, b); NTL_OPT_RETURN(ZZXY, x); }

inline ZZXY operator*(const ZZX& a, const ZZXY& b)
   { ZZXY x; mul(x, a, b); NTL_OPT_RETURN(ZZXY, x); }

inline ZZXY operator*(const ZZ& a, const ZZXY& b)
   { ZZXY x; mul(x, a, b); NTL_OPT_RETURN(ZZXY, x); }

inline ZZXY operator*(long a, const ZZXY& b)
   { ZZXY x; mul(x, a, b); NTL_OPT_RETURN(ZZXY, x); }

inline ZZXY& operator*=(ZZXY& x, const ZZXY& b)
   { mul(x, x, b); return x; }

inline ZZXY& operator*=(ZZXY& x, const ZZX& b)
   { mul(x, x, b); return x; }

inline ZZXY& operator*=(ZZXY& x, const ZZ& b)
   { mul(x, x, b); return x; }

inline ZZXY& operator*=(ZZXY& x, long b)
   { mul(x, x, b); return x; }



/*************************************************************

                      Division

**************************************************************/



// "plain" versions
void PlainPseudoDivRem(ZZXY& q, ZZXY& r, const ZZXY& a, const ZZXY& b);
void PlainPseudoDiv(ZZXY& q, const ZZXY& a, const ZZXY& b);
void PlainPseudoRem(ZZXY& r, const ZZXY& a, const ZZXY& b);


inline void PseudoDivRem(ZZXY& q, ZZXY& r, const ZZXY& a, const ZZXY& b)
// performs pseudo-division:  computes q and r
// with deg(r) < deg(b), and LeadCoeff(b)^(deg(a)-deg(b)+1) a = b q + r.
// current implementation always defaults to "plain"

   { PlainPseudoDivRem(q, r, a, b); }

inline void PseudoDiv(ZZXY& q, const ZZXY& a, const ZZXY& b)

   { PlainPseudoDiv(q, a, b); }

inline void PseudoRem(ZZXY& r, const ZZXY& a, const ZZXY& b)

   { PlainPseudoRem(r, a, b); }

inline ZZXY PseudoDiv(const ZZXY& a, const ZZXY& b)
   { ZZXY x; PseudoDiv(x, a, b); NTL_OPT_RETURN(ZZXY, x); }

inline ZZXY PseudoRem(const ZZXY& a, const ZZXY& b)
   { ZZXY x; PseudoRem(x, a, b); NTL_OPT_RETURN(ZZXY, x); }


#ifndef NTL_TRANSITION

void DivRem(ZZXY& q, ZZXY& r, const ZZXY& a, const ZZXY& b);

void div(ZZXY& q, const ZZXY& a, const ZZXY& b);
void div(ZZXY& q, const ZZXY& a, const ZZX& b);
void div(ZZXY& q, const ZZXY& a, const ZZ& b);
void div(ZZXY& q, const ZZXY& a, long b);

void rem(ZZXY& r, const ZZXY& a, const ZZXY& b);

inline ZZXY operator/(const ZZXY& a, const ZZXY& b)
   { ZZXY x; PseudoDiv(x, a, b); NTL_OPT_RETURN(ZZXY, x); }

inline ZZXY operator/(const ZZXY& a, const ZZ& b)
{ ZZXY x; div(x, a, b); NTL_OPT_RETURN(ZZXY, x); }

inline ZZXY operator/(const ZZXY& a, const ZZX& b)
   { ZZXY x; div(x, a, b); NTL_OPT_RETURN(ZZXY, x); }

inline ZZXY operator/(const ZZXY& a, long b)
   { ZZXY x; div(x, a, b); NTL_OPT_RETURN(ZZXY, x); }

inline ZZXY& operator/=(ZZXY& x, const ZZ& b)
   { div(x, x, b); return x; }

inline ZZXY& operator/=(ZZXY& x, const ZZX& b)
   { div(x, x, b); return x; }

inline ZZXY& operator/=(ZZXY& x, long b)
   { div(x, x, b); return x; }

inline ZZXY& operator/=(ZZXY& x, const ZZXY& b)
   { PseudoDiv(x, x, b); return x; }


inline ZZXY operator%(const ZZXY& a, const ZZXY& b)
   { ZZXY x; PseudoRem(x, a, b); NTL_OPT_RETURN(ZZXY, x); }

inline ZZXY& operator%=(ZZXY& x, const ZZXY& b)
   { PseudoRem(x, x, b); return x; }

#endif


long divide(ZZXY& q, const ZZXY& a, const ZZXY& b);
// if b | a, sets q = a/b and returns 1; otherwise returns 0

long divide(const ZZXY& a, const ZZXY& b);


long divide(ZZXY& q, const ZZXY& a, const ZZX& b);
// if b | a, sets q = a/b and returns 1; otherwise returns 0

long divide(const ZZXY& a, const ZZX& b);
// if b | a, returns 1; otherwise returns 0

long divide(ZZXY& q, const ZZXY& a, const ZZ& b);
long divide(const ZZXY& a, const ZZ& b);


//single-precision versions
long divide(ZZXY& q, const ZZXY& a, long b);
long divide(const ZZXY& a, long b);



void content(ZZX& d, const ZZXY& f);
// d = content of f, sign(d) == sign(LeadCoeff(f))

inline ZZX content(const ZZXY& f)
   { ZZX x; content(x, f); NTL_OPT_RETURN(ZZX, x); }



// vectors

#if 0
NTL_vector_decl(ZZX,vec_ZZX)

NTL_eq_vector_decl(ZZX,vec_ZZX)

NTL_io_vector_decl(ZZX,vec_ZZX)
#endif

NTL_CLOSE_NNS

#endif
